package dev.fyloz.colorrecipesexplorer.config.security

import com.fasterxml.jackson.module.kotlin.jacksonObjectMapper
import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.account.UserDetails
import dev.fyloz.colorrecipesexplorer.model.account.UserLoginRequest
import dev.fyloz.colorrecipesexplorer.model.account.UserOutputDto
import dev.fyloz.colorrecipesexplorer.model.account.toAuthorities
import dev.fyloz.colorrecipesexplorer.service.users.JwtService
import dev.fyloz.colorrecipesexplorer.service.users.UserDetailsService
import dev.fyloz.colorrecipesexplorer.utils.addCookie
import io.jsonwebtoken.ExpiredJwtException
import org.springframework.security.authentication.AuthenticationManager
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken
import org.springframework.security.core.Authentication
import org.springframework.security.core.context.SecurityContextHolder
import org.springframework.security.web.authentication.UsernamePasswordAuthenticationFilter
import org.springframework.security.web.authentication.www.BasicAuthenticationFilter
import org.springframework.web.util.WebUtils
import javax.servlet.FilterChain
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse

const val authorizationCookieName = "Authorization"
const val defaultGroupCookieName = "Default-Group"
val blacklistedJwtTokens = mutableListOf<String>()  // Not working, move to a cache or something

class JwtAuthenticationFilter(
    private val authManager: AuthenticationManager,
    private val jwtService: JwtService,
    private val securityProperties: CreSecurityProperties,
    private val updateUserLoginTime: (Long) -> Unit
) : UsernamePasswordAuthenticationFilter() {
    private var debugMode = false

    init {
        setFilterProcessesUrl("/api/login")
        debugMode = "debug" in environment.activeProfiles
    }

    override fun attemptAuthentication(request: HttpServletRequest, response: HttpServletResponse): Authentication {
        val loginRequest = jacksonObjectMapper().readValue(request.inputStream, UserLoginRequest::class.java)
        return authManager.authenticate(UsernamePasswordAuthenticationToken(loginRequest.id, loginRequest.password))
    }

    override fun successfulAuthentication(
        request: HttpServletRequest,
        response: HttpServletResponse,
        chain: FilterChain,
        auth: Authentication
    ) {
        val userDetails = auth.principal as UserDetails
        val token = jwtService.buildJwt(userDetails)

        response.addHeader(authorizationCookieName, "Bearer $token")
        response.addCookie(authorizationCookieName, "Bearer$token") {
            httpOnly = true
            sameSite = true
            secure = !debugMode
            maxAge = securityProperties.jwtDuration / 1000
        }

        updateUserLoginTime(userDetails.user.id)
    }
}

class JwtAuthorizationFilter(
    private val jwtService: JwtService,
    authenticationManager: AuthenticationManager,
    private val userDetailsService: UserDetailsService
) : BasicAuthenticationFilter(authenticationManager) {
    override fun doFilterInternal(request: HttpServletRequest, response: HttpServletResponse, chain: FilterChain) {
        fun tryLoginFromBearer(): Boolean {
            val authorizationCookie = WebUtils.getCookie(request, authorizationCookieName)
            // Check for an authorization token cookie or header
            val authorizationToken = if (authorizationCookie != null)
                authorizationCookie.value
            else
                request.getHeader(authorizationCookieName)

            // An authorization token is valid if it starts with "Bearer", is not expired and is not blacklisted
            if (authorizationToken != null && authorizationToken.startsWith("Bearer") && authorizationToken !in blacklistedJwtTokens) {
                val authenticationToken = getAuthentication(authorizationToken) ?: return false
                SecurityContextHolder.getContext().authentication = authenticationToken
                return true
            }
            return false
        }

        fun tryLoginFromDefaultGroupCookie() {
            val defaultGroupCookie = WebUtils.getCookie(request, defaultGroupCookieName)
            if (defaultGroupCookie != null) {
                val authenticationToken = getAuthenticationToken(defaultGroupCookie.value)
                SecurityContextHolder.getContext().authentication = authenticationToken
            }
        }

        if (!tryLoginFromBearer())
            tryLoginFromDefaultGroupCookie()
        chain.doFilter(request, response)
    }

    private fun getAuthentication(token: String): UsernamePasswordAuthenticationToken? {
        return try {
            val user = jwtService.parseJwt(token.replace("Bearer", ""))
            getAuthenticationToken(user)
        } catch (_: ExpiredJwtException) {
            null
        }
    }

    private fun getAuthenticationToken(user: UserOutputDto) =
        UsernamePasswordAuthenticationToken(user.id, null, user.permissions.toAuthorities())

    private fun getAuthenticationToken(userId: Long): UsernamePasswordAuthenticationToken? = try {
        val userDetails = userDetailsService.loadUserById(userId)
        UsernamePasswordAuthenticationToken(userDetails.username, null, userDetails.authorities)
    } catch (_: NotFoundException) {
        null
    }

    private fun getAuthenticationToken(userId: String) =
        getAuthenticationToken(userId.toLong())
}
