package dev.fyloz.colorrecipesexplorer.model.account

import dev.fyloz.colorrecipesexplorer.SpringUserDetails
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.EntityDto
import dev.fyloz.colorrecipesexplorer.model.Model
import org.hibernate.annotations.Fetch
import org.hibernate.annotations.FetchMode
import org.springframework.security.crypto.bcrypt.BCryptPasswordEncoder
import org.springframework.security.crypto.password.PasswordEncoder
import java.time.LocalDateTime
import javax.persistence.*
import javax.validation.constraints.NotBlank
import javax.validation.constraints.Size

private const val VALIDATION_PASSWORD_LENGTH = "Must contains at least 8 characters"

@Entity
@Table(name = "user")
data class User(
    @Id
    override val id: Long,

    @Column(name = "first_name")
    val firstName: String = "",

    @Column(name = "last_name")
    val lastName: String = "",

    val password: String = "",

    @Column(name = "default_group_user")
    val isDefaultGroupUser: Boolean = false,

    @Column(name = "system_user")
    val isSystemUser: Boolean = false,

    @ManyToOne
    @JoinColumn(name = "group_id")
    @Fetch(FetchMode.SELECT)
    var group: Group? = null,

    @Enumerated(EnumType.STRING)
    @ElementCollection(fetch = FetchType.EAGER)
    @CollectionTable(name = "user_permission", joinColumns = [JoinColumn(name = "user_id")])
    @Column(name = "permission")
    @Fetch(FetchMode.SUBSELECT)
    val permissions: MutableSet<Permission> = mutableSetOf(),

    @Column(name = "last_login_time")
    var lastLoginTime: LocalDateTime? = null
) : Model {
    val flatPermissions: Set<Permission>
        get() = permissions
            .flatMap { it.flat() }
            .filter { !it.deprecated }
            .toMutableSet()
            .apply {
                if (group != null) this.addAll(group!!.flatPermissions)
            }
}

open class UserSaveDto(
    val id: Long,

    @field:NotBlank
    val firstName: String,

    @field:NotBlank
    val lastName: String,

    @field:NotBlank
    @field:Size(min = 8, message = VALIDATION_PASSWORD_LENGTH)
    val password: String,

    val groupId: Long?,

    @Enumerated(EnumType.STRING)
    val permissions: MutableSet<Permission> = mutableSetOf()
) : EntityDto<User>

open class UserUpdateDto(
    val id: Long,

    @field:NotBlank
    val firstName: String?,

    @field:NotBlank
    val lastName: String?,

    val groupId: Long?,

    @Enumerated(EnumType.STRING)
    val permissions: Set<Permission>?
) : EntityDto<User>

data class UserOutputDto(
    override val id: Long,
    val firstName: String,
    val lastName: String,
    val group: Group?,
    val permissions: Set<Permission>,
    val explicitPermissions: Set<Permission>,
    val lastLoginTime: LocalDateTime?
) : Model

data class UserLoginRequest(val id: Long, val password: String)

data class UserDetails(val user: User) : SpringUserDetails {
    override fun getPassword() = user.password
    override fun getUsername() = user.id.toString()
    override fun getAuthorities() = user.flatPermissions.toAuthorities()

    override fun isAccountNonExpired() = true
    override fun isAccountNonLocked() = true
    override fun isCredentialsNonExpired() = true
    override fun isEnabled() = true
}

// ==== DSL ====
fun user(
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = "password",
    isDefaultGroupUser: Boolean = false,
    isSystemUser: Boolean = false,
    group: Group? = null,
    permissions: MutableSet<Permission> = mutableSetOf(),
    lastLoginTime: LocalDateTime? = null,
    op: User.() -> Unit = {}
) = User(
    id,
    firstName,
    lastName,
    password,
    isDefaultGroupUser,
    isSystemUser,
    group,
    permissions,
    lastLoginTime
).apply(op)

fun user(
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    plainPassword: String = "password",
    isDefaultGroupUser: Boolean = false,
    isSystemUser: Boolean = false,
    group: Group? = null,
    permissions: MutableSet<Permission> = mutableSetOf(),
    lastLoginTime: LocalDateTime? = null,
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    op: User.() -> Unit = {}
) = User(
    id,
    firstName,
    lastName,
    passwordEncoder.encode(plainPassword),
    isDefaultGroupUser,
    isSystemUser,
    group,
    permissions,
    lastLoginTime
).apply(op)

fun userSaveDto(
    passwordEncoder: PasswordEncoder = BCryptPasswordEncoder(),
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    password: String = passwordEncoder.encode("password"),
    groupId: Long? = null,
    permissions: MutableSet<Permission> = mutableSetOf(),
    op: UserSaveDto.() -> Unit = {}
) = UserSaveDto(id, firstName, lastName, password, groupId, permissions).apply(op)

fun userUpdateDto(
    id: Long = 0L,
    firstName: String = "firstName",
    lastName: String = "lastName",
    groupId: Long? = null,
    permissions: MutableSet<Permission> = mutableSetOf(),
    op: UserUpdateDto.() -> Unit = {}
) = UserUpdateDto(id, firstName, lastName, groupId, permissions).apply(op)

// ==== Extensions ====
fun Set<Permission>.toAuthorities() =
    this.map { it.toAuthority() }.toMutableSet()

fun User.toOutputDto() =
    UserOutputDto(
        this.id,
        this.firstName,
        this.lastName,
        this.group,
        this.flatPermissions,
        this.permissions,
        this.lastLoginTime
    )

// ==== Exceptions ====
private const val USER_NOT_FOUND_EXCEPTION_TITLE = "User not found"
private const val USER_ALREADY_EXISTS_EXCEPTION_TITLE = "User already exists"
private const val USER_EXCEPTION_ERROR_CODE = "user"

fun userIdNotFoundException(id: Long) =
    NotFoundException(
        USER_EXCEPTION_ERROR_CODE,
        USER_NOT_FOUND_EXCEPTION_TITLE,
        "An user with the id $id could not be found",
        id
    )

fun userIdAlreadyExistsException(id: Long) =
    AlreadyExistsException(
        USER_EXCEPTION_ERROR_CODE,
        USER_ALREADY_EXISTS_EXCEPTION_TITLE,
        "An user with the id $id already exists",
        id
    )

fun userFullNameAlreadyExistsException(firstName: String, lastName: String) =
    AlreadyExistsException(
        USER_EXCEPTION_ERROR_CODE,
        USER_ALREADY_EXISTS_EXCEPTION_TITLE,
        "An user with the name '$firstName $lastName' already exists",
        "$firstName $lastName",
        "fullName"
    )
