package dev.fyloz.colorrecipesexplorer.service.users

import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.module.kotlin.readValue
import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.model.account.User
import dev.fyloz.colorrecipesexplorer.model.account.UserDetails
import dev.fyloz.colorrecipesexplorer.model.account.UserOutputDto
import dev.fyloz.colorrecipesexplorer.model.account.toOutputDto
import dev.fyloz.colorrecipesexplorer.utils.base64encode
import dev.fyloz.colorrecipesexplorer.utils.toDate
import io.jsonwebtoken.Jwts
import io.jsonwebtoken.jackson.io.JacksonDeserializer
import io.jsonwebtoken.jackson.io.JacksonSerializer
import org.springframework.stereotype.Service
import java.time.Instant
import java.util.*

const val jwtClaimUser = "user"

interface JwtService {
    /** Build a JWT token for the given [userDetails]. */
    fun buildJwt(userDetails: UserDetails): String

    /** Build a JWT token for the given [user]. */
    fun buildJwt(user: User): String

    /** Parses a user from the given [jwt] token. */
    fun parseJwt(jwt: String): UserOutputDto
}

@Service
class JwtServiceImpl(
    val objectMapper: ObjectMapper,
    val securityProperties: CreSecurityProperties
) : JwtService {
    private val secretKey by lazy {
        securityProperties.jwtSecret.base64encode()
    }

    private val jwtBuilder by lazy {
        Jwts.builder()
            .serializeToJsonWith(JacksonSerializer<Map<String, *>>(objectMapper))
            .signWith(secretKey)
    }

    private val jwtParser by lazy {
        Jwts.parserBuilder()
            .deserializeJsonWith(JacksonDeserializer<Map<String, *>>(objectMapper))
            .setSigningKey(secretKey)
            .build()
    }

    override fun buildJwt(userDetails: UserDetails) =
        buildJwt(userDetails.user)

    override fun buildJwt(user: User): String =
        jwtBuilder
            .setSubject(user.id.toString())
            .setExpiration(getCurrentExpirationDate())
            .claim(jwtClaimUser, user.serialize())
            .compact()

    override fun parseJwt(jwt: String): UserOutputDto =
        with(
            jwtParser.parseClaimsJws(jwt)
                .body.get(jwtClaimUser, String::class.java)
        ) {
            objectMapper.readValue(this)
        }

    private fun getCurrentExpirationDate(): Date =
        Instant.now()
            .plusSeconds(securityProperties.jwtDuration)
            .toDate()

    private fun User.serialize(): String =
        objectMapper.writeValueAsString(this.toOutputDto())
}
