package dev.fyloz.colorrecipesexplorer.service.users

import dev.fyloz.colorrecipesexplorer.config.security.blacklistedJwtTokens
import dev.fyloz.colorrecipesexplorer.model.account.*
import dev.fyloz.colorrecipesexplorer.model.validation.or
import dev.fyloz.colorrecipesexplorer.repository.UserRepository
import dev.fyloz.colorrecipesexplorer.service.AbstractExternalModelService
import dev.fyloz.colorrecipesexplorer.service.ExternalModelService
import org.springframework.context.annotation.Lazy
import org.springframework.context.annotation.Profile
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.stereotype.Service
import org.springframework.web.util.WebUtils
import java.time.LocalDateTime
import javax.servlet.http.HttpServletRequest

interface UserService :
    ExternalModelService<User, UserSaveDto, UserUpdateDto, UserOutputDto, UserRepository> {
    /** Check if an [User] with the given [firstName] and [lastName] exists. */
    fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean

    /** Gets the user with the given [id]. */
    fun getById(id: Long, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): User

    /** Gets all users which have the given [group]. */
    fun getByGroup(group: Group): Collection<User>

    /** Gets the default user of the given [group]. */
    fun getDefaultGroupUser(group: Group): User

    /** Save a default group user for the given [group]. */
    fun saveDefaultGroupUser(group: Group)

    /** Updates de given [entity]. **/
    fun update(entity: User, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): User

    /** Updates the last login time of the user with the given [userId]. */
    fun updateLastLoginTime(userId: Long, time: LocalDateTime = LocalDateTime.now()): User

    /** Updates the password of the user with the given [id]. */
    fun updatePassword(id: Long, password: String): User

    /** Adds the given [permission] to the user with the given [userId]. */
    fun addPermission(userId: Long, permission: Permission): User

    /** Removes the given [permission] from the user with the given [userId]. */
    fun removePermission(userId: Long, permission: Permission): User

    /** Logout an user. Add the authorization token of the given [request] to the blacklisted tokens. */
    fun logout(request: HttpServletRequest)
}

@Service
@Profile("!emergency")
class UserServiceImpl(
    userRepository: UserRepository,
    @Lazy val groupService: GroupService,
) : AbstractExternalModelService<User, UserSaveDto, UserUpdateDto, UserOutputDto, UserRepository>(
    userRepository
),
    UserService {
    override fun idNotFoundException(id: Long) = userIdNotFoundException(id)
    override fun idAlreadyExistsException(id: Long) = userIdAlreadyExistsException(id)

    override fun User.toOutput() = this.toOutputDto()

    override fun existsByFirstNameAndLastName(firstName: String, lastName: String): Boolean =
        repository.existsByFirstNameAndLastName(firstName, lastName)

    override fun getAll(): Collection<User> =
        super.getAll().filter { !it.isSystemUser && !it.isDefaultGroupUser }

    override fun getById(id: Long): User =
        getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)

    override fun getById(id: Long, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): User =
        super.getById(id).apply {
            if (ignoreSystemUsers && isSystemUser || ignoreDefaultGroupUsers && isDefaultGroupUser)
                throw idNotFoundException(id)
        }

    override fun getByGroup(group: Group): Collection<User> =
        repository.findAllByGroup(group).filter {
            !it.isSystemUser && !it.isDefaultGroupUser
        }

    override fun getDefaultGroupUser(group: Group): User =
        repository.findByIsDefaultGroupUserIsTrueAndGroupIs(group)

    override fun save(entity: UserSaveDto): User =
        save(with(entity) {
            user(
                id = id,
                firstName = firstName,
                lastName = lastName,
                plainPassword = password,
                isDefaultGroupUser = false,
                isSystemUser = false,
                group = if (groupId != null) groupService.getById(groupId) else null,
                permissions = permissions
            )
        })

    override fun save(entity: User): User {
        if (existsById(entity.id))
            throw userIdAlreadyExistsException(entity.id)
        if (existsByFirstNameAndLastName(entity.firstName, entity.lastName))
            throw userFullNameAlreadyExistsException(entity.firstName, entity.lastName)
        return super<AbstractExternalModelService>.save(entity)
    }

    override fun saveDefaultGroupUser(group: Group) {
        save(
            user(
                id = 1000000L + group.id!!,
                firstName = group.name,
                lastName = "User",
                plainPassword = group.name,
                group = group,
                isDefaultGroupUser = true
            )
        )
    }

    override fun updateLastLoginTime(userId: Long, time: LocalDateTime): User {
        val user = getById(userId, ignoreDefaultGroupUsers = true, ignoreSystemUsers = false)
        user.lastLoginTime = time
        return update(
            user,
            ignoreDefaultGroupUsers = true,
            ignoreSystemUsers = false
        )
    }

    override fun update(entity: UserUpdateDto): User {
        val persistedUser by lazy { getById(entity.id) }
        return update(with(entity) {
            User(
                id = id,
                firstName = firstName or persistedUser.firstName,
                lastName = lastName or persistedUser.lastName,
                password = persistedUser.password,
                isDefaultGroupUser = false,
                isSystemUser = false,
                group = if (entity.groupId != null) groupService.getById(entity.groupId) else persistedUser.group,
                permissions = permissions?.toMutableSet() ?: persistedUser.permissions,
                lastLoginTime = persistedUser.lastLoginTime
            )
        })
    }

    override fun update(entity: User): User =
        update(entity, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)

    override fun update(entity: User, ignoreDefaultGroupUsers: Boolean, ignoreSystemUsers: Boolean): User {
        with(repository.findByFirstNameAndLastName(entity.firstName, entity.lastName)) {
            if (this != null && id != entity.id)
                throw userFullNameAlreadyExistsException(entity.firstName, entity.lastName)
        }

        return super.update(entity)
    }

    override fun updatePassword(id: Long, password: String): User {
        val persistedUser = getById(id, ignoreDefaultGroupUsers = true, ignoreSystemUsers = true)
        return super.update(with(persistedUser) {
            user(
                id,
                firstName,
                lastName,
                plainPassword = password,
                isDefaultGroupUser,
                isSystemUser,
                group,
                permissions,
                lastLoginTime
            )
        })
    }

    override fun addPermission(userId: Long, permission: Permission): User =
        super.update(getById(userId).apply { permissions += permission })

    override fun removePermission(userId: Long, permission: Permission): User =
        super.update(getById(userId).apply { permissions -= permission })

    override fun logout(request: HttpServletRequest) {
        val authorizationCookie = WebUtils.getCookie(request, "Authorization")
        if (authorizationCookie != null) {
            val authorizationToken = authorizationCookie.value
            if (authorizationToken != null && authorizationToken.startsWith("Bearer")) {
                blacklistedJwtTokens.add(authorizationToken)
            }
        }
    }
}
