package dev.fyloz.colorrecipesexplorer.utils

import org.apache.pdfbox.pdmodel.PDDocument
import org.apache.pdfbox.pdmodel.PDPage
import org.apache.pdfbox.pdmodel.PDPageContentStream
import org.apache.pdfbox.pdmodel.font.PDFont
import org.apache.pdfbox.pdmodel.font.PDType1Font
import org.springframework.core.io.ByteArrayResource
import java.io.ByteArrayOutputStream

val PDF_DEFAULT_FONT: PDType1Font = PDType1Font.HELVETICA
val PDF_DEFAULT_FONT_BOLD: PDType1Font = PDType1Font.HELVETICA_BOLD
const val PDF_DEFAULT_FONT_SIZE = 42f
val PDF_DASH_LINE_PATTERN = floatArrayOf(4f)

/** Creates a [PdfContainer] and apply the given [block].  */
fun pdf(block: PdfDocument.() -> Unit = {}) =
    PdfDocument().apply { block() }

/** Creates a [PdfContainer] in the given [PdfDocument] and apply the given [block]. If a [container] is given, the receiver of the block will be a clone of it. */
fun PdfDocument.container(container: PdfContainer = PdfContainer(), block: PdfContainer.() -> Unit) {
    this.containers += PdfContainer(container).apply(block)
}

/** Creates a [PdfText] with the given [text] in the given [PdfContainer] and apply the given [block]. */
fun PdfContainer.text(text: String, block: PdfText.() -> Unit) {
    this.texts += PdfText(text = text).apply(block)
}

fun PdfDocument.toByteArrayResource(): ByteArrayResource = PDDocument().use { document ->
    val page = PDPage()

    document.addPage(page)

    fun PDPageContentStream.drawText(text: PdfText, y: Float) {
        val font = if (text.bold) fontBold else font
        val textWidth = font.getStringWidth(text.text) / 1000 * text.fontSize
        val textX = (page.mediaBox.width - textWidth) / 2f

        beginText()
        newLineAtOffset(textX, y)
        setFont(font, text.fontSize)
        showText(text.text)
        endText()
    }

    fun PDPageContentStream.drawDashLine(y: Float) {
        moveTo(0f, y)
        lineTo(page.mediaBox.width, y)
        setLineDashPattern(PDF_DASH_LINE_PATTERN, 0f)
        stroke()
    }

    fun PDPageContentStream.drawContainer(container: PdfContainer, y: Float, height: Float) {
        var textY = y

        if (container.centeredVertically) {
            val textsHeight = container.texts
                .map { it.fontSize + it.marginTop }
                .reduce { acc, textHeight -> acc + textHeight }
            textY -= (height - textsHeight) / 2f
        }

        if (container.drawContainerBottom) {
            this.drawDashLine(y - height)
        }

        container.texts.forEach { text ->
            textY -= text.fontSize + text.marginTop
            this.drawText(text, textY)
        }
    }

    PDPageContentStream(document, page).use {
        var containerY = page.mediaBox.height

        val computedSizeContainerCount = containers
            .filter { it.height < 0 }
            .count()
        val computedSizeContainersHeight = containerY / computedSizeContainerCount

        containers.forEach { container ->
            val height = if (container.height < 0)
                computedSizeContainersHeight
            else
                container.height

            it.drawContainer(container, containerY, height)

            containerY -= height
        }
    }

    ByteArrayOutputStream().use {
        document.save(it)
        ByteArrayResource(it.toByteArray())
    }
}

data class PdfDocument(
    var font: PDFont = PDF_DEFAULT_FONT,
    var fontBold: PDFont = PDF_DEFAULT_FONT_BOLD,
    val containers: MutableList<PdfContainer> = mutableListOf()
)

data class PdfContainer(
    var height: Float = -1f,
    var centeredVertically: Boolean = false,
    var drawContainerBottom: Boolean = false,
    val texts: MutableList<PdfText> = mutableListOf()
) {
    constructor(container: PdfContainer) : this(
        container.height,
        container.centeredVertically,
        container.drawContainerBottom,
        container.texts
    )
}

data class PdfText(
    var text: String = "Text",
    var bold: Boolean = false,
    var fontSize: Float = PDF_DEFAULT_FONT_SIZE,
    var marginTop: Float = 0f
)
