package dev.fyloz.colorrecipesexplorer.service

import com.fasterxml.jackson.module.kotlin.jacksonObjectMapper
import com.fasterxml.jackson.module.kotlin.readValue
import dev.fyloz.colorrecipesexplorer.config.properties.CreSecurityProperties
import dev.fyloz.colorrecipesexplorer.model.account.UserDetails
import dev.fyloz.colorrecipesexplorer.model.account.UserOutputDto
import dev.fyloz.colorrecipesexplorer.model.account.toOutputDto
import dev.fyloz.colorrecipesexplorer.model.account.user
import dev.fyloz.colorrecipesexplorer.service.users.JwtServiceImpl
import dev.fyloz.colorrecipesexplorer.service.users.jwtClaimUser
import dev.fyloz.colorrecipesexplorer.utils.base64encode
import dev.fyloz.colorrecipesexplorer.utils.isAround
import io.jsonwebtoken.Jwts
import io.jsonwebtoken.jackson.io.JacksonDeserializer
import io.mockk.spyk
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test
import java.time.Instant
import kotlin.test.assertEquals
import kotlin.test.assertTrue

class JwtServiceTest {
    private val objectMapper = jacksonObjectMapper()
    private val securityProperties = CreSecurityProperties().apply {
        jwtSecret = "XRRm7OflmFuCrOB2Xvmfsercih9DCKom"
        jwtDuration = 1000000L
    }
    private val jwtParser by lazy {
        Jwts.parserBuilder()
            .deserializeJsonWith(JacksonDeserializer<Map<String, *>>(objectMapper))
            .setSigningKey(securityProperties.jwtSecret.base64encode())
            .build()
    }

    private val jwtService = spyk(JwtServiceImpl(objectMapper, securityProperties))

    private val user = user()
    private val userOutputDto = user.toOutputDto()

    // buildJwt()

    private fun withParsedUserOutputDto(jwt: String, test: (UserOutputDto) -> Unit) {
        val serializedUser = jwtParser.parseClaimsJws(jwt)
            .body.get(jwtClaimUser, String::class.java)

        test(objectMapper.readValue(serializedUser))
    }

    @Test
    fun `buildJwt(userDetails) returns jwt string with valid user`() {
        val userDetails = UserDetails(user)

        val builtJwt = jwtService.buildJwt(userDetails)

        withParsedUserOutputDto(builtJwt) { parsedUser ->
            assertEquals(user.toOutputDto(), parsedUser)
        }
    }

    @Test
    fun `buildJwt() returns jwt string with valid user`() {
        val builtJwt = jwtService.buildJwt(user)

        withParsedUserOutputDto(builtJwt) { parsedUser ->
            assertEquals(user.toOutputDto(), parsedUser)
        }
    }

    @Test
    fun `buildJwt() returns jwt string with valid subject`() {
        val builtJwt = jwtService.buildJwt(user)
        val jwtSubject = jwtParser.parseClaimsJws(builtJwt).body.subject

        assertEquals(user.id.toString(), jwtSubject)
    }

    @Test
    fun `buildJwt() returns jwt with valid expiration date`() {
        val jwtExpectedExpirationDate = Instant.now().plusSeconds(securityProperties.jwtDuration)

        val builtJwt = jwtService.buildJwt(user)
        val jwtExpiration = jwtParser.parseClaimsJws(builtJwt)
            .body.expiration.toInstant()

        // Check if it's between 1 second
        assertTrue { jwtExpiration.isAround(jwtExpectedExpirationDate) }
    }

    // parseJwt()

    @Test
    fun `parseJwt() returns expected user`() {
        val jwt = jwtService.buildJwt(user)
        val parsedUser = jwtService.parseJwt(jwt)

        assertEquals(userOutputDto, parsedUser)
    }
}
